@@ -1,17 +1,18 @@
-from typing import NamedTuple, Optional, Union
-from types import FunctionType
+from typing import NamedTuple, Optional, Union, Deque, Tuple, Callable, Any
 from select import select
 from time import sleep, time
 from collections import deque
 from hashlib import sha256
 from binascii import a2b_hex
 from Cryptodome.Cipher import AES
+from Cryptodome.Cipher._mode_gcm import GcmMode
 from struct import Struct
 from io import BytesIO, SEEK_END
 from socket import socket
 import socket as s
 import threading
 import logging
+import atexit
 import ecdsa
 import os
 
@@ -24,7 +25,7 @@ CONTROL_PSH = 0b00000010  # Push data immediately
 CONTROL_EOF = 0b00000100  # end of file
 CONTROL_BCT = 0b00001000  # broadcast
 CONTROL_RTM = 0b00010000  # ask retransmission
-CONTROL_MTU = 0b00100000  # fix MTU size
+# CONTROL_MTU = 0b00100000  # fix MTU size
 CONTROL_FIN = 0b01000000  # fin
 FLAG_NAMES = {
     0b00000000: "---",
@@ -34,7 +35,7 @@ FLAG_NAMES = {
     CONTROL_PSH | CONTROL_EOF: "PSH+EOF",
     CONTROL_BCT: "BCT",
     CONTROL_RTM: "RTM",
-    CONTROL_MTU: "MTU",
+    # CONTROL_MTU: "MTU",
     CONTROL_FIN: "FIN",
 }
 WINDOW_MAX_SIZE = 32768  # 32kb
@@ -54,6 +55,11 @@ S_SEND_PUBLIC_KEY = b'\x01'
 S_SEND_SHARED_KEY = b'\x02'
 S_ESTABLISHED = b'\x03'
 
+# typing
+_Address = Tuple[Any, ...]
+_WildAddress = Union[_Address, str, bytes]
+_BroadcastHook = Callable[['Packet'], None]
+
 
 class CycInt(int):
     """
@@ -116,27 +122,26 @@ class Packet(NamedTuple):
     time: float  # unix time (double)
     data: bytes  # data body
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "Packet({} seq:{} retry:{} time:{} data:{}b)".format(
             FLAG_NAMES.get(self.control), self.sequence,
             self.retry, round(self.time, 2), len(self.data))
 
 
-def bin2packet(b: bytes):
+def bin2packet(b: bytes) -> 'Packet':
     c, seq, r, t = packet_struct.unpack_from(b)
     return Packet(c, CycInt(seq), r, t, b[packet_struct.size:])
 
 
-def packet2bin(p: Packet):
+def packet2bin(p: Packet) -> bytes:
     # log.debug("s>> %s", p)
     return packet_struct.pack(p.control, int(p.sequence), p.retry, p.time) + p.data
 
 
-def get_formal_address_format(address: tuple, family=s.AF_INET) -> tuple:
+def get_formal_address_format(address: _WildAddress, family: int = s.AF_INET) -> _Address:
     """tuple of ipv4/6 correct address format"""
-    assert isinstance(address, tuple)
+    assert isinstance(address, tuple), "cannot recognize bytes or str format"
     for _, _, _, _, addr in s.getaddrinfo(str(address[0]), int(address[1]), family, s.SOCK_STREAM):
-        # domain name resolve
         return addr
     else:
         raise ConnectionError("not found correct ip format of {}".format(address))
@@ -144,13 +149,12 @@ def get_formal_address_format(address: tuple, family=s.AF_INET) -> tuple:
 
 class SecureReliableSocket(socket):
     __slots__ = [
-        "timeout", "span", "address", "shared_key",
-        "mut_auto_fix", "mtu_size", "mtu_multiple",
+        "timeout", "span", "address", "shared_key", "mtu_size",
         "sender_seq", "sender_buffer", "sender_signal", "sender_buffer_lock",
         "receiver_seq", "receiver_buffer", "receiver_signal", "receiver_buffer_lock",
-        "broadcast_hook_fnc", "loss", "established"]
+        "broadcast_hook_fnc", "loss", "try_connect", "established"]
 
-    def __init__(self, family=s.AF_INET, timeout=21.0, span=3.0):
+    def __init__(self, family: int = s.AF_INET, timeout: float = 21.0, span: float = 3.0) -> None:
         """
         :param family: socket type AF_INET or AF_INET6
         :param timeout: auto socket close by the time passed (sec)
@@ -160,14 +164,12 @@ class SecureReliableSocket(socket):
         # inner params
         self.timeout = timeout
         self.span = span
-        self.address = tuple()
-        self.shared_key = b''
-        self.mut_auto_fix = False  # set automatic best MUT size
+        self.address: _Address = None
+        self.shared_key: bytes = None
         self.mtu_size = 0  # 1472b
-        self.mtu_multiple = 1  # 1 to 4096
         # sender params
         self.sender_seq = CycInt(1)  # next send sequence
-        self.sender_buffer = deque()
+        self.sender_buffer: Deque[Packet] = deque()
         self.sender_signal = threading.Event()  # clear when buffer is empty
         self.sender_buffer_lock = threading.Lock()
         # receiver params
@@ -176,17 +178,59 @@ class SecureReliableSocket(socket):
         self.receiver_signal = threading.Event()
         self.receiver_buffer_lock = threading.Lock()
         # broadcast hook
-        self.broadcast_hook_fnc: Optional[FunctionType] = None
+        self.broadcast_hook_fnc: Optional[_BroadcastHook] = None
         # status
         self.loss = 0
+        self.try_connect = False
         self.established = False
 
-    def connect(self, address: tuple) -> None:
-        """UDP hole punching & get shared key"""
+    def connect(self, address: _WildAddress) -> None:
+        """
+        `connection()` establishment process
+        ====
+        ```uml
+                          ┌─┐                                            ┌─┐
+                          │A│                                            │B│
+                          └┬┘                                            └┬┘
+                           │              udp-hole-punching               │
+                           │────────────────────────────────────────────> │
+                           │                                              │
+                           │           udp-hole-punching (fail)           │
+                           │ X<───────────────────────────────────────────│
+                           │                                              │
+                           │               send B's publicKey             │
+                           │ <────────────────────────────────────────────│
+                           │                                              │
+                           │ send sharedKey (encrypted by sharedPoint)    │ ╔═══════════════╗
+                           │───────────────────────────────────────────────>║ B established ║
+                           │                                              │ ╚═══════════════╝
+        ╔════════════════╗ │ send establish flag (encrypted by sharedKey) │
+        ║ A established  ║<───────────────────────────────────────────────│
+        ╚════════════════╝ │                                              │
+                          ┌┴┐                                            ┌┴┐
+                          │A│                                            │B│
+                          └─┘                                            └─┘
+        ```
+
+        note
+        ----
+        * only one of two hole-punching is success in most case.
+        * when both hole-punching is success, use high priority side's sharedKey, but is rare case.
+        * when both hole-punching is fail, you can't use UDP-hole-punching method in your network.
+        * sharedPoint is calculated by multiply secretKey with publicKey.
+        * sharedKey is random 256bit bytes, don't use sharedPoint as sharedKey.
+        """
         assert not self.established, "already established"
+        assert not self.is_closed, "already closed socket"
+        assert not self.try_connect, "already try to connect"
+
+        # start communication (only once you can try)
+        self.try_connect = True
+
+        # bind socket address
         self.address = address = get_formal_address_format(address, self.family)
         address_copy = list(address)
-        address_copy[0] = ""  # bind global address
+        address_copy[0] = ""  # the port is CLOSE_WAIT state if this raise OSError
         self.bind(tuple(address_copy))
         log.debug("try to communicate with {}".format(address))
 
@@ -198,8 +242,12 @@ class SecureReliableSocket(socket):
         punch_msg = b"udp hole punching"
         self.sendto(S_HOLE_PUNCHING + punch_msg + select_curve.name.encode(), address)
 
-        # my secret key
+        # my secret & public key
         my_sk: Optional[ecdsa.SigningKey] = None
+        my_pk: Optional[ecdsa.VerifyingKey] = None
+
+        # other's public key
+        other_pk: Optional[ecdsa.VerifyingKey] = None
 
         check_msg = b"success hand shake"
         for _ in range(int(self.timeout / self.span)):
@@ -235,7 +283,20 @@ class SecureReliableSocket(socket):
                 elif stage == S_SEND_SHARED_KEY:
                     # 4. decrypt shared key & send hello msg
                     encrypted_data = data.decode().split("+")
-                    other_pk = ecdsa.VerifyingKey.from_string(a2b_hex(encrypted_data[0]), select_curve)
+                    if other_pk is None:
+                        other_pk = ecdsa.VerifyingKey.from_string(a2b_hex(encrypted_data[0]), select_curve)
+                    else:
+                        log.debug("need to check priority because already get others's pk")
+                        my_pri = sha256(my_pk.to_string() + other_pk.to_string()).digest()
+                        other_pri = sha256(other_pk.to_string() + my_pk.to_string()).digest()
+                        # check 256bit int as big-endian
+                        if my_pri < other_pri:
+                            log.debug("my priority is LOW and over write my shared key by other's")
+                        elif other_pri < my_pri:
+                            log.debug("my priority is HIGH and ignore this command")
+                            continue
+                        else:
+                            raise ConnectionError("my and other's key is same, this means you connect to yourself")
                     if my_sk is None:
                         raise ConnectionError("not found my_sk")
                     shared_point = my_sk.privkey.secret_multiplier * other_pk.pubkey.point
@@ -273,6 +334,9 @@ class SecureReliableSocket(socket):
         threading.Thread(target=self._backend, daemon=True).start()
         self.established = True
 
+        # auto exit when program closed
+        atexit.register(self.close)
+
     def _find_mut_size(self) -> int:
         """confirm by submit real packet"""
         wait = 0.05
@@ -311,13 +375,12 @@ class SecureReliableSocket(socket):
     def _backend(self) -> None:
         """reorder sequence & fill output buffer"""
         temporary = dict()
-        retransmit_packets = deque()
-        retransmitted = deque(maxlen=16)
-        broadcast_packets = deque(maxlen=16)
+        retransmit_packets: Deque[Packet] = deque()
+        retransmitted: Deque[float] = deque(maxlen=16)
+        broadcast_packets: Deque[Packet] = deque(maxlen=16)
         last_packet: Optional[Packet] = None
         last_receive_time = time()
         last_ack_time = time()
-        last_mtu_fix_time = time()
 
         while not self.is_closed:
             r, _w, _x = select([self], [], [], self.span)
@@ -349,17 +412,6 @@ class SecureReliableSocket(socket):
                 self.sendto(self._encrypt(packet2bin(p)), self.address)
                 break
 
-            # fix MTU size more bigger
-            if self.mut_auto_fix and self.timeout < time() - last_mtu_fix_time:
-                try:
-                    new_mtu = self.mtu_size * min(self.mtu_multiple + 1, 4096)
-                    size = 16 * (new_mtu // 16 - 1) - packet_struct.size - 1
-                    p = Packet(CONTROL_MTU, CycInt(new_mtu), 0, time(), b'#' * size)
-                    self.sendto(self._encrypt(packet2bin(p)), self.address)
-                except s.error:
-                    pass  # Message too long
-                last_mtu_fix_time = time()
-
             # received packet
             if r:
                 try:
@@ -409,30 +461,6 @@ class SecureReliableSocket(socket):
                                 self.sendto(self._encrypt(packet2bin(re_packet)), self.address)
                                 retransmitted.append(packet.time)
                                 break
-                    # too big? fix MTU smaller
-                    if self.span < time() - last_mtu_fix_time:
-                        if 1 < len(retransmitted):
-                            ave = (time() - retransmitted[0]) / len(retransmitted)
-                            if 1 < self.mtu_multiple and ave < self.timeout:
-                                try:
-                                    new_mtu = self.mtu_size * max(self.mtu_multiple - 1, 1)
-                                    size = 16 * (new_mtu // 16 - 1) - packet_struct.size - 1
-                                    p = Packet(CONTROL_MTU, CycInt(new_mtu), 0, time(), b'#' * size)
-                                    self.sendto(self._encrypt(packet2bin(p)), self.address)
-                                except s.error:
-                                    pass  # Message too long
-                        last_mtu_fix_time = time()
-                    continue
-
-                # receive MTU fix packet
-                if packet.control & CONTROL_MTU:
-                    if len(packet.data) == 0:
-                        # receive response: update to new MTU
-                        self.mtu_multiple = int(packet.sequence) // self.mtu_size
-                    else:
-                        # send response: success new MTU
-                        p = Packet(CONTROL_MTU, packet.sequence, 0, time(), b'')
-                        self.sendto(self._encrypt(packet2bin(p)), self.address)
                     continue
 
                 # broadcast packet
@@ -527,7 +555,7 @@ class SecureReliableSocket(socket):
         # close
         self.close()
 
-    def _push_receive_buffer(self, data: bytes):
+    def _push_receive_buffer(self, data: bytes) -> None:
         """just append new data to buffer"""
         with self.receiver_buffer_lock:
             pos = self.receiver_buffer.tell()
@@ -541,9 +569,15 @@ class SecureReliableSocket(socket):
 
     def get_window_size(self) -> int:
         """maximum size of data you can send at once"""
-        return self.mtu_size * self.mtu_multiple - 32 - packet_struct.size
+        return self.mtu_size - 32 - packet_struct.size
 
-    def send(self, data: memoryview, flags=0) -> int:
+    def send(self, data: bytes, flags: int = 0) -> int:
+        """over write low-level method for compatibility"""
+        assert flags == 0, "unrecognized flags"
+        self.sendall(data)
+        return len(data)
+
+    def _send(self, data: memoryview) -> int:
         """warning: row-level method"""
         if not self.established:
             raise ConnectionAbortedError('disconnected')
@@ -574,17 +608,16 @@ class SecureReliableSocket(socket):
                 break
         return send_size
 
-    def sendall(self, data: Union[bytes, memoryview], flags=0) -> None:
+    def sendall(self, data: bytes, flags: int = 0) -> None:
         """high-level method, use this instead of send()"""
-        assert flags == 0, "flags are disabled"
+        assert flags == 0, "unrecognized flags"
         if not self._send_buffer_is_full():
             self.sender_signal.set()
         send_size = 0
-        if isinstance(data, bytes):
-            data = memoryview(data)
+        data = memoryview(data)
         while send_size < len(data):
             if self.sender_signal.wait(self.timeout):
-                send_size += self.send(data[send_size:])
+                send_size += self._send(data[send_size:])
 
     def broadcast(self, data: bytes) -> None:
         """broadcast data (do not check reach)"""
@@ -598,8 +631,8 @@ class SecureReliableSocket(socket):
         with self.sender_buffer_lock:
             self.sendto(self._encrypt(packet2bin(packet)), self.address)
 
-    def recv(self, buflen=1024, flags=0) -> bytes:
-        assert flags == 0, "?"
+    def recv(self, buflen: int = 1024, flags: int = 0) -> bytes:
+        assert flags == 0, "unrecognized flags"
         timeout = self.gettimeout()
         while not self.is_closed:
             if not self.established:
@@ -632,7 +665,7 @@ class SecureReliableSocket(socket):
 
     def _encrypt(self, data: bytes) -> bytes:
         """encrypt by AES-GCM (more secure than CBC mode)"""
-        cipher = AES.new(self.shared_key, AES.MODE_GCM)
+        cipher: 'GcmMode' = AES.new(self.shared_key, AES.MODE_GCM)  # type: ignore
         # warning: Don't reuse nonce
         enc, tag = cipher.encrypt_and_digest(data)
         # output length = 16bytes + 16bytes + N(=data)bytes
@@ -640,7 +673,7 @@ class SecureReliableSocket(socket):
 
     def _decrypt(self, data: bytes) -> bytes:
         """decrypt by AES-GCM (more secure than CBC mode)"""
-        cipher = AES.new(self.shared_key, AES.MODE_GCM, nonce=data[:16])
+        cipher: 'GcmMode' = AES.new(self.shared_key, AES.MODE_GCM, nonce=data[:16])  # type: ignore
         # ValueError raised when verify failed
         return cipher.decrypt_and_verify(data[32:], data[16:32])
 
@@ -648,20 +681,22 @@ class SecureReliableSocket(socket):
     def is_closed(self) -> bool:
         if self.fileno() == -1:
             self.established = False
+            atexit.unregister(self.close)
             return True
         return False
 
     def close(self) -> None:
         if self.established:
+            self.established = False
             p = Packet(CONTROL_FIN, CYC_INT0, 0, time(), b'closed')
             self.sendto(self._encrypt(packet2bin(p)), self.address)
             sleep(0.001)
             super().close()
-            self.established = False
+            atexit.unregister(self.close)
             self.receiver_signal.set()
 
 
-def find_ecdhe_curve(curve_name):
+def find_ecdhe_curve(curve_name: str) -> ecdsa.curves.Curve:
     for curve in ecdsa.curves.curves:
         if curve.name == curve_name:
             return curve
@@ -669,7 +704,7 @@ def find_ecdhe_curve(curve_name):
         raise ConnectionError("unknown curve {}".format(curve_name))
 
 
-def get_mtu_linux(family, host) -> int:
+def get_mtu_linux(family: int, host: str) -> int:
     """MTU on Linux"""
     with socket(family, s.SOCK_DGRAM) as sock:
         sock.connect((host, 0))
@@ -679,7 +714,7 @@ def get_mtu_linux(family, host) -> int:
         return sock.getsockopt(s.IPPROTO_IP, IP_MTU)
 
 
-def main():
+def main() -> None:
     """for test"""
     import sys, random
     remote_host = sys.argv[1]
@@ -699,7 +734,7 @@ def main():
     sock.connect((remote_host, port))
     log.debug("connect success! mtu=%d", sock.mtu_size)
 
-    def listen():
+    def listen() -> None:
         size, start = 0, time()
         while True:
             r = sock.recv(8192)
@@ -714,7 +749,7 @@ def main():
             # log.debug("recv %d %d", size, len(r))
         log.debug("closed receive")
 
-    def sending():
+    def sending() -> None:
         while msglen:
             sock.sendall(b'start!'+os.urandom(msglen)+b'success!')  # +14
             log.debug("send now! loss=%d time=%d", sock.loss, int(time()))
@@ -723,7 +758,7 @@ def main():
                 log.debug("send broadcast!")
             sleep(20)
 
-    def broadcast_hook(packet: Packet):
+    def broadcast_hook(packet: Packet) -> None:
         log.debug("find you!!! (%s)", packet)
 
     sock.broadcast_hook_fnc = broadcast_hook
